/*
 * This file is part of MinimalGS, which is a GameScript for OpenTTD
 * Copyright (C) 2012  Leif Linse
 *
 * MinimalGS is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * MinimalGS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MinimalGS; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

require("root_class.nut");

/** Import SuperLib for GameScript **/
if (!("SuperLib" in _LibServerGS_private_const_root_class)) {
	import("util.superlib", "SuperLib", 36);
	_LibServerGS_private_const_root_class.SuperLib <- SuperLib; // import into the GS too
} else {
	// Detect if loaded SuperLib is >= version 36.
	if (!("HasWorldGenBug" in SuperLib.Helper)) {
		Log.Error("LibServerGS want to have SuperLib 36 or later. An older version was detected.");
		throw false;
	}
}

require("exception.nut");
require("version.nut");

class LibServerGS extends GSController 
{
	// Users of this library should not access any member variables directly.
	// Use the public interface below.

	// An array with classes at root level that are exposed via Admin Port.
	static _allowed_root_prefix = ["GS"];
	static _allowed_root_exact_match = ["Story", "SuperLib"]; // (Story exist for backwards compatibility with v1-2)

	// Our own uniqueid used for errors displayed using GSGoal.Question
	// If remote Admin Port client uses this uniqueid a warning will be issued.
	static _own_uniqueid = 0;

	// Setting for if errors should be displayed to all companies using
	// GSGoal.Question
	static _display_errors = false;

	// reference to root class containing all APIs, loaded libraries etc.
	// this "variable" should be never be changed.
	// See also root_class.nut
	static _const_root_class = _LibServerGS_private_const_root_class;

	// this exception is used internally in the library when a call/symbol
	// cannot be resolved. See exception.nut for implementation.
	static _exception = _LibServerGS_private_exception;

	// Log messages are prefixed with this string
	static _pre_log = "LibServerGS: ";

	constructor()
	{
	}

	//// Public interface: ////

	// main method:
	function ReceiveEvent(event);

	// may be useful for some:
	function AddAllowedRootPrefix(new_prefix);
	function AddAllowedRootExactMatch(new_match);
	function GetOwnUniqueid();
	function Eval(call, args);
	function GetSymbol(str);

	/* Also this library reads the GS setting LSGS_show_error_dialogs:
	 * 
	 * enables/disables displaying of error messages to all companies using
	 * GSGoal.Question. If enabled, and a missing symbol is called or other
	 * parsing error occurs, a blue window is shown to all companies.
	 *
	 * This setting may be useful to debug Admin Port clients.
	 *
	 * A value of 1 => enabled, otherwise disabled.
	 */

	/* And, as it uses SuperLib.Log for logging, read the docs of SuperLib
	 * if you want to use its log level setting. (either in your settings
	 * dialog or if you by code want to filter what to display)
	 */

	///////////////////////////
}

/*
 * Add a allowed prefix for classes/methods at the root level
 * that will be exposed via the Admin Port.
 * By default anything that begins with GS is accepted
 */
function LibServerGS::AddAllowedRootPrefix(new_prefix)
{
	if (typeof(new_prefix) != "string") return false;
	LibServerGS._allowed_root_exact_match.append(new_prefix);
	return true;
}

/*
 * Add a allowed exact match for classes/methods at the root level
 * that will be exposed via the Admin Port.
 * By default anything that exactly match Story and SuperLib is exposed
 */
function LibServerGS::AddAllowedRootExactMatch(new_match)
{
	if (typeof(new_match) != "string") return false;
	LibServerGS._allowed_root_exact_match.append(new_match);
	return true;
}

/*
 * Get uniqueid used for displaying error messages
 */
function LibServerGS::GetOwnUniqueid()
{
	return LibServerGS._own_uniqueid;
}

/*
 * Call this method when an event has been received to allow
 * this library to gain access to the event.
 *
 * If you do not call this method, no incoming Admin Port
 * packets will be received by the library.
 */
function LibServerGS::ReceiveEvent(event)
{
	local ev_type = event.GetEventType();
	switch (ev_type) {
		/* Incoming message from Admin port */
		case GSEvent.ET_ADMIN_PORT: {
			local admin_event = GSEventAdminPort.Convert(event);
			local data = admin_event.GetObject();

			LibServerGS.ReceiveAdminPortMessage(data);
			break;
		}

		/* If a button was pressed, send an event unless our own internal uniqueid was used. */
		case GSEvent.ET_GOAL_QUESTION_ANSWER: {
			local question_event = GSEventGoalQuestionAnswer.Convert(event);
			if (question_event.GetUniqueID() == LibServerGS._own_uniqueid) break;
			local message = {
				action = "event",
				event_type = "ET_GOAL_QUESTION_ANSWER",
				unique_id = question_event.GetUniqueID(),
				company = question_event.GetCompany(),
				button = question_event.GetButton(),
			}
			GSAdmin.Send(message);
		}
		

		// other events ...
	}

}

/*
 * This method is called by ReceiveEvent when
 * an incoming message from the admin port has been
 * received.
 * @param data Squirrel object constructed from JSON data
 * @return void
 */
function LibServerGS::ReceiveAdminPortMessage(data) 
{
	if (data == null) throw LibServerGS._exception("Received JSON data is null");
	if (typeof(data) != "table") throw LibServerGS._exception("Received JSON root structure is not a table");
	if (!data.rawin("action")) throw LibServerGS._exception("\"action\" field missing in the JSON table");

	// Create basic response table
	local response = {
		action=data.action,
		result=null
	}

	// Send back message number if given in data packet
	if (data.rawin("number")) response.number <- data.number;

	try {
		// Which action was requested?
		if (data.action == "call") {
			if (!data.rawin("method")) throw LibServerGS._exception("\"method\" field missing in the JSON table");
			if (!data.rawin("args")) throw LibServerGS._exception("\"args\" field missing in the JSON table");

			// If data.companymode exist, it give a company ID which should be used
			// to construct a GSCompanyMode in the scope where the requested method
			// will be called.
			local cm = null;
			if (data.rawin("companymode") && typeof(data.companymode) == "integer") {
				cm = GSCompanyMode(data.companymode);
			}

			// If data.testmode exist, create a GSTestMode instance.
			local tm = null;
			if (data.rawin("testmode")) {
				tm = GSTestMode();
			}

			response.result = LibServerGS.Eval(data.method, data.args);

			// Log call + result
			local s = data.method + "(";
			local arg_str = "";
			foreach (arg in data.args) {
				if (arg_str != "") arg_str += ", ";
				arg_str += arg;
			}
			s += arg_str;
			s += ")";
			if (cm != null) s += " [company:" + data.companymode + "]";
			if (tm != null) s += " [testmode]";
			s += " => " + response.result;
			SuperLib.Log.Info(LibServerGS._pre_log + s, SuperLib.Log.LVL_SUB_DECISIONS);
		} else if (data.action == "ping") {
			response.result = "ping";
			SuperLib.Log.Info(LibServerGS._pre_log + "Received a ping from admin port");
		} else {
			throw LibServerGS._exception("Unknown action");
		}

		// Send response
		GSAdmin.Send(response);
	} catch(e) {
		SuperLib.Log.Info(LibServerGS._pre_log + "--- to here ---", SuperLib.Log.LVL_INFO);
		SuperLib.Log.Warning(LibServerGS._pre_log + "An error occurred while processing incoming Admin Port command due to unknown data format.", SuperLib.Log.LVL_INFO);
		response.error <- true;
		GSAdmin.Send(response);

		if (GSController.GetSetting("LSGS_show_error_dialogs") == 1) {
			GSGoal.Question(LibServerGS._own_uniqueid, GSCompany.COMPANY_INVALID, "" + e._str, GSGoal.QT_ERROR, GSGoal.BUTTON_OK);
		}
	}

}

/*
 * Evaluate function/constructor call and return the
 * result value.
 * @param call A string containing the name of method to call including class. Example: "GSTown.GetName"
 * @param args An array containing the arguments to a method. The type of each argument is important. See implementation.
 * @return the raw return value
 */
function LibServerGS::Eval(call, args)
{
	local func = LibServerGS.GetSymbol(call);
	if (typeof(func) != "function" && typeof(func) != "class") throw LibServerGS._exception("Function to call (\"" + call + "\") is not a function nor constructor");
	local arg_values = [];
	local arg_num = 0; // argument number
	foreach(arg in args) {
		arg_num++;

		local arg_value = null;

		// Check type of argument string
		if (typeof(arg) == "string") {
			// A string can either be a literal string or a reference to an enum in the API
			local str = LibServerGS.Trim(arg);
			if (str[0] == '"' || str[0] == "'"[0]) {
				// Literal string
				local end_char = str[0];
				if (str[str.len()-1] != end_char) throw LibServerGS._exception("Unmatched " + end_char);
				arg_value = str.slice(1, str.len() - 1); // remove leading and terminating ".
			} else {
				// reference to API enum

				// Some enums are used as flags that can be joined using bitwise OR. Therefore
				// specifically support the OR operator. (but no other operator)
				local or_parts = SuperLib.Helper.SplitString("|", str);
				if (or_parts.len() == 0) {
					or_parts.append(str);
				}

				// Iterate over all referenced symbols and join their value using OR.
				arg_value = 0;
				foreach(or_part in or_parts) {
					local sym = LibServerGS.GetSymbol(or_part);
					arg_value = arg_value | sym.symbol;
				}
			}
		} else if (typeof(arg) == "integer" || typeof(arg) == "boolean") {
			arg_value = arg;
		} else {
			throw LibServerGS._exception("Unsupported argument type: " + typeof(arg) + " of argument \"" + arg_num + "\"");
		}
			
		arg_values.append(arg_value);
	}

	if (call == "GSGoal.Question" && arg_values[0] == LibServerGS._own_uniqueid) {
		local warning1 = "You should not use the uniqueid " + LibServerGS._own_uniqueid + " for GSGoal.Question because it is reserved for internal use (error dialogs like this one). ";
		local warning2 = "If you use the reserved uniqueid, you will not receive the button click event";
		if (GSController.GetSetting("LSGS_show_error_dialogs") == 1) {
			GSGoal.Question(LibServerGS._own_uniqueid, GSCompany.COMPANY_INVALID, warning1 + warning2, GSGoal.QT_WARNING, GSGoal.BUTTON_OK);
		}
		SuperLib.Log.Warning(LibServerGS._pre_log + warning1, SuperLib.Log.LVL_INFO);
		SuperLib.Log.Info(LibServerGS._pre_log + warning2, SuperLib.Log.LVL_INFO);
	}

	local result = SuperLib.Helper.CallFunction(func, arg_values);

	// If the result is an instance it is assumed to be a sub class of GSList and need to 
	// be converted to an Squirrel array so that it can be send back as JSON.
	if (typeof(result) == "instance") {
		SuperLib.Log.Info(LibServerGS._pre_log + "Result is a class, try to convert it to an squirrel array", SuperLib.Log.LVL_DEBUG);
		local sq_array = [];
		foreach (val, _ in result) {
			sq_array.append(val);
		}
		result = sq_array;
	}

	return result;
}

function LibServerGS::Trim(str) {
	// Trim start
	local pos = str.find(" ");
	while (pos == 0) {
		str = str.slice(1);
		pos = str.find(" ");
	}

	// Trim end
	pos = SuperLib.Helper.FindFromEnd(" ", str);
	while (pos == str.len() - 1) {
		str = str.slice(0, str.len() - 1);
		pos = SuperLib.Helper.FindFromEnd(" ", str);
	}

	return str;
}

/*
 * This method will lookup a reference/pointer to the given symbol.
 * @note only references to symbol allowed by the prefix filter are returned
 * @param str A string containing the symbol to lookup. Examples: "GSTown.GetName" or "SuperLib.Helper.GetPAXCargo" or "GSTownList"
 * @return reference/pointer to symbol (which can be used to call the method)
 */
function LibServerGS::GetSymbol(str) {
	local parts = SuperLib.Helper.SplitString(".", LibServerGS.Trim(str));
	if (parts.len() == 0) throw LibServerGS._exception("Empty symbol \"" + str + "\"");
	local curr_level = LibServerGS._const_root_class;
	for (local i = 0; i < parts.len(); i++) {
		if (i == 0 && !LibServerGS.AcceptedRootSymbol(parts[i])) continue;
		if (i == 0 && parts[i] == "Story") { /* Special case for calling SuperLib.Story from root-level for compatibility with version 1 and 2 of ServerGS */
			curr_level = LibServerGS.GetSymbolInScope(parts[i], SuperLib);
		} else {
			curr_level = LibServerGS.GetSymbolInScope(parts[i], curr_level);
		}
		if (curr_level == null) {
			throw LibServerGS._exception("Couldn't find requested symbol: " + str);
		}
	}
	return curr_level;
}

/*
 * Is given symbol at name accepted by the root level prefix filter?
 * @param str Symbol name
 * @return true if accepted, otherwise false
 */
function LibServerGS::AcceptedRootSymbol(str)
{
	foreach (accepted in LibServerGS._allowed_root_prefix) {
		if (str.slice(0, accepted.len()) == accepted) {
			return true;
		}
	}
	foreach (accepted in LibServerGS._allowed_root_exact_match) {
		if (str == accepted) {
			return true;
		}
	}
}

 /*
  * This method will look for a given symbol directly within a given scope (class)
  * @param symbol_name The name of the symbol to lookup
  * @param scope A reference to the class/scope where a symbol with the name symbol_name is supposed to be found.
  * @return a reference to the symbol or null (if symbol couldn't be found)
  */
function LibServerGS::GetSymbolInScope(symbol_name, scope)
{
	foreach (item, value in scope)
	{
		if (symbol_name == item) {
			return value;
		}
	}
	return null;
}

/*
 * Count number of exposed symbols
 * @param scope Gives the scope from where to look for symbols. This 
 *        parameter is used for recursive calls. Leave it as null to get
 *        all symbols.
 * @return Number of exposed symbols.
 */
function LibServerGS::CountExposedSymbols(scope = null) 
{
	// Iterates over all symbols and return leaf symbol count.
	// Note that this method do not count the number of symbols in the
	// Story module twice unless the GS include it in global scope.
	local count = 0;
	local root = scope == null;
	if (root) scope = LibServerGS._const_root_class;

	foreach(name, symbol in scope) {
		if (root && !LibServerGS.AcceptedRootSymbol(name)) continue;

		if (typeof(symbol) == "class") { 
			count += 1 + // constructor
					LibServerGS.CountExposedSymbols(symbol)
		} else /*if (typeof(symbol) == "function" || 
				typeof(symbol) == "integer" ||
				typeof(symbol) == "bool" ||
				typeof(symbol) == "float" ||
				typeof(symbol) == "table" ||
				typeof(symbol) == "array" ||
				typeof(symbol) == "string")*/ {
			// function / flag
			count++;
		}
	}
	
	return count;
}

// Initialization:
SuperLib.Log.Info("LibServerGS: " + LibServerGS.CountExposedSymbols() + " symbols are exposed to Admin Port clients. LibServerGS version: " + SELF_VERSION, SuperLib.Log.LVL_INFO);

